import padre as p
import neural as nl
import json,os,shutil,copy

''' padre v2.0 design:

subject
    |- dsets(label,exp,...)
    |- meta('demographics','clinical','behavioral','neuropsych')
        # list available options:
    |- sessions, experiments, labels, tags

dset
    |- complete
    |- meta('eprime','scan_sheet')
    |- info #dset_info object
    |- experiment, label, tags, session
'''

class PrefixDict(dict):
    def __init__(self,*args):
        dict.__init__(self,*args)
        self.prefix = None
    
    def __getitem__(self,key,raw=False):
        if self.prefix!=None and raw==False:
            return self.prefix + dict.__getitem__(self,key)
        return dict.__getitem__(self,key)

class Dset(object):
    '''object to contain dataset filename and meta-data
    
    Calling :meth:`subject.dsets` with return a list of these objects. They can be treated as 
    strings most of the time, because they will return the absolute location of the dataset
    filename when cast as strings. If you want to explicitly convert them into the file location
    you can manually cast them as a string, ``str(dset)``. To return only the filename,
    with no directory information, you can call :meth:`__str__` with the argument ``False``
    
    Other information contained in the objects include:
    
    :complete:      Whether this is a complete, usable dataset
    :date:          Date this was acquired
    :experiment:    Experiment in which this was aquired
    :label:         Kind of dataset (e.g., ``anatomy``)
    :tags:          Tags of session this was acquired during
    :session:       Label of session this was acquired during
    :md5:           md5 checksum of dataset file (used for checking for data corruption)
    :meta:          dictionary of meta data associated with this dataset (e.g., ``eprime`` or ``eprime-txt``)
    :info:          object containing information from a call to ``3dinfo`` (for more information see the neural library)'''
    
    def __init__(self,subject,session,dset_fname,label=None,complete=True,md5=None,meta={}):
        self._dset_fname = dset_fname
        self.complete = complete
        self.meta = PrefixDict(meta)
        self.meta.prefix = os.path.join(p.sessions_dir(subject),session)
        self.md5 = md5
        
        self.date = subject._sessions[session]['date'] if 'date' in subject._sessions[session] else None
        self.experiment = subject._sessions[session]['experiment'] if 'experiment' in subject._sessions[session] else None
        self.label = label
        if self.label == None:
            try:
                for l in subject._sessions[session]['labels']:
                    for d in subject._sessions[session]['labels'][l]:
                        if d['filename'] == dset_fname:
                            self.label = l
                            raise StopIteration
            except StopIteration:
                pass
        self.tags = subject._sessions[session]['tags'] if 'tags' in subject._sessions[session] else None
        self.session = session
        
        self._info = None
        self._subject = subject
    
    def __abspath__(self):
        '''return the absolute path the the dataset'''
        return os.path.join(p.sessions_dir(self._subject),self.session,self._dset_fname)
    
    def __dict__(self):
        return {
            'filename': self._dset_fname,
            'complete': self.complete,
            'md5': self.md5,
            'meta': self.meta
        }
    
    @classmethod
    def from_dict(cls,subject,session,dict_source):
        dset = Dset(subject,session,dict_source['filename'])
        dset.complete = dict_source['complete'] if 'complete' in dict_source else True
        dset.md5 = dict_source['md5'] if 'md5' in dict_source else None
        dset.meta = PrefixDict(dict_source['meta'])
        dset.meta.prefix = os.path.join(p.sessions_dir(subject),session)
        return dset
    
    def __str__(self,absolute=True):
        if absolute:
            return self.__abspath__()
        else:
            return self._dset_fname
    
    def __repr__(self):
        return self._dset_fname
    
    def __getattribute__(self,name):
        if name=='info':
            if self._info==None:
                try:
                    self._info = nl.afni.dset_info(self.__abspath__())
                except:
                    pass
            return self._info
        else:
            return object.__getattribute__(self,name)

class Subject(object):
    '''abstract container for subject information
    
    If cast as string(``str(subject)``), will return subject id. Subject objects can be obtained by 
    calling the search function :meth:`padre.subjects` or by loading directly using
    :meth:`padre.load`
    
    .. autoinstanceattribute:: Subject.include
        :annotation:
    
    .. autoinstanceattribute:: Subject.sessions
        :annotation:
        
    .. autoinstanceattribute:: Subject.labels
        :annotation:
        
    .. autoinstanceattribute:: Subject.experiments
        :annotation:
    
    .. autoinstanceattribute:: Subject.meta
        :annotation:
    
    .. autoinstanceattribute:: Subject.notes
        :annotation:
    
    .. automethod:: Subject.load
    
    .. automethod:: Subject.dsets
    
    '''
    def __init__(self,subject_id,initial_data={}):
        self._subject_id = subject_id
        #: If ``False``, this subject will be excluded from standard analyses
        self.include = initial_data['include'] if 'include' in initial_data else True  
        #: Free text notes on subject
        self.notes = initial_data['notes'] if 'notes' in initial_data else ''
        
        #: List of session names
        self.sessions = []
        #: List of labels used by this subject
        self.labels = []
        #: List of experiments this subject has data for
        self.experiments = []
        self._dsets = []
        
        self._sessions = initial_data['sessions'] if 'sessions' in initial_data else {}
        for sess in self._sessions:
            if 'unverified' in self._sessions[sess] and not p._include_all:
                continue
            self.sessions.append(sess)
            for label in self._sessions[sess]['labels']:
                self.labels.append(label)
                for dset in self._sessions[sess]['labels'][label]:
                    self._dsets.append(Dset(self,sess,dset['filename'],label,dset['complete'],dset['meta']))
        self.labels = list(set(self.labels))
        self.experiments = list(set(self.labels))
        
        self.meta = p.ForgivingDict.copy_nested_dict(initial_data['meta']) if 'meta' in initial_data else p.ForgivingDict()
        '''Dictionary of meta-data
        Standard entries include: ``clinical``, ``demo`` (i.e., demographic), ``neuropsych``, and ``behavioral``'''
    
    @classmethod
    def load(cls,subject_id):
        ''' returns a subject object initialized using JSON file
        If no subject can be loaded, will print an error and return ``None``
        
        This method is imported to the root-level, and can be called as :meth:``padre.load``'''
        subject_id = str(subject_id).rstrip('/')
        json_file = p.subject_json(subject_id)
        try:
            with open(json_file) as f:
                return Subject(subject_id,json.loads(f.read()))
        except (ValueError,IOError):
            p.error('Could not load valid JSON file for subject %s' % subject_id)
            return None
    
    def __dict__(self):
        return {
            'include': self.include,
            'notes': self.notes,
            'meta': self.meta,
            'sessions': self._sessions
        }
             
    def save(self,json_file=None):
        ''' save current state to JSON file '''
        if json_file==None:
            json_file = p.subject_json(self._subject_id)
        save_dict = self.__dict__()
        with open(json_file,'w') as f:
            f.write(json.dumps(save_dict, sort_keys=True, indent=2))
    
    def init_directories(self):
        ''' create directories that these scripts expect on the disk '''
        for d in [
                p.subject_dir(self),
                p.raw_dir(self),
                p.sessions_dir(self)
            ]:
            if not os.path.exists(d):
                os.makedirs(d)   
    
    def __repr__(self):
        return "subject(%s)" % (self._subject_id)
    
    def __str__(self):
        return self._subject_id
    
    def dsets(self,label=None,experiment=None,session=None,tags=None,include_all=False):
        '''returns list of :class:`Dset` objects matching the given criteria
        
        This is the primary method for accessing the datasets. This does not return strings, but :class:`Dset` objects.
        Datasets can be searched using the following parameters:
        
            :label:         type of dataset (e.g., ``anatomy`` or a task name)
            :experiment:    name of experiment this dataset belongs to
            :session:       name of session this dataset was collected in
            :tags:          arbitrary tag for session (e.g., ``preop`` or ``session_5``)
            
        By default, datasets that are tagged as ``unverified`` or ``incomplete`` are not returned. To return all
        datasets, call with ``include_all=True``'''
        if experiment==None and p._global_experiment:
            experiment = p._global_experiment
        return_dsets = []
        if session:
            include_sessions = [session]
        else:
            include_sessions = self._sessions.keys()
        if include_all==False:
            include_sessions = [x for x in include_sessions if p._include_all or 'unverified' not in self._sessions[x] and ('include' not in self._sessions[x] or self._sessions[x]['include']==True)]
        for sess in include_sessions:
            if sess in self._sessions:
                if tags:
                    if isinstance(tags,basestring):
                        tags = [tags]
                    if 'tags' not in self._sessions[sess] or not all([tag in self._sessions[sess]['tags'] for tag in tags]):
                        continue
                if experiment:
                    if self._sessions[sess]['experiment']!=experiment:
                        continue
                if label:
                    include_labels = [label]
                else:
                    include_labels = self._sessions[sess]['labels']
                for label in include_labels:
                    if label in self._sessions[sess]['labels']:
                        return_dsets += [Dset.from_dict(self,sess,dset) for dset in self._sessions[sess]['labels'][label] if include_all or dset['complete']]
        return return_dsets


tasks = set()
experiments = set()
tags = set()
_all_subjects = {}
_indexed_and_loaded_all_subjects = False

def _index_one_subject(subject_id):
    _all_subjects[subject_id] = Subject.load(subject_id)
    for session in _all_subjects[subject_id]._sessions:
        session_dict = _all_subjects[subject_id]._sessions[session]
        if 'experiment' in session_dict and session_dict['experiment']!='' and session_dict['experiment']!=None:
            experiments.add(session_dict['experiment'])
        if 'labels' in session_dict and session_dict['labels']!='' and session_dict['labels']!=None:
            [tasks.add(x) for x in session_dict['labels']]
        if 'tags' in session_dict and session_dict['tags']!=[] and session_dict['tags']!=None:
            for tag in session_dict['tags']: 
                tags.add(tag)

def _index_all_subjects(load_all=False):
    global _all_subjects
    if os.path.exists(p.data_dir):
        for subject_id in os.listdir(p.data_dir):
            if os.path.exists(p.subject_json(subject_id)):
                if subject_id not in _all_subjects:
                    _all_subjects[subject_id] = None
                if load_all and _all_subjects[subject_id]==None:
                    _index_one_subject(subject_id)
    if load_all:
        _indexed_and_loaded_all_subjects = True

_index_all_subjects()

def subjects(experiment=None,label=None,tags=None,only_included=True):
    '''returns a list of subject objects for all subjects with valid JSON files
    
    :experiment:    only return subjects that have a scan for the given experiment
    :label:         only return subjects who have datasets with that label
    :tags:          only return subjects who have a session with given tag
    
    :only_included: if True (the default), will exclude any subjects with ``subject.include``
                    set to False
    
    Using this list, it is easy to filter again to find a specific subset of subjects. 
    
    For example, for all the subjects for the experiment ``Exp1``::
    
        subjects = padre.subjects('Exp1')
    
    To find all subjects who contain a run called ``rest``::
    
        subjects = padre.subjects(label='rest')
    
    To get a list of all subjects who have had a left ATL surgery::

    	subjects = [s for s in padre.subjects() if s.meta['clinical']['surgery_type']=='left atl']
    
    '''
    if not _indexed_and_loaded_all_subjects:
        _index_all_subjects(True)
    if experiment==None and p._global_experiment:
        experiment = p._global_experiment
    all_subjs = _all_subjects.values()
    if label:
        all_subjs = [x for x in all_subjs if len(x.dsets(label))]
    if experiment:
        all_subjs = [x for x in all_subjs if experiment in [x._sessions[y]['experiment'] for y in x._sessions if 'experiment' in x._sessions[y]]]
    if tags:
        if isinstance(tags,basestring):
            tags = [tags]
        all_subjs = [x for x in all_subjs if all([tag in nl.flatten([x._sessions[y]['tags'] for y in x._sessions if 'tags' in x._sessions[y]]) for tag in tags])]
    if only_included and not p._include_all:
        all_subjs = [x for x in all_subjs if x.include]
    
    return all_subjs

def subject_exists(subj_id):
    return os.path.exists(p.subject_json(subj_id))

class DatabaseConsistencyError(Exception):
    pass

def _files_exist():
    '''make sure all the files that should be there, are still there'''
    for s in p.subjects():
        for dset in s.dsets:
            if not os.path.exists(dset):
                raise DatabaseConsistencyError("dataset is missing: %s" % dset)
        for sess in s.sessions:
            if 'scan_sheets' in sess:
                if not os.path.exists(os.path.join(p.sessions_dir(s),sess,sess['scan_sheets'])):
                    raise DatabaseConsistencyError("scan sheets PDF is missing: %s" % sess['scan_sheets'])

def sanity_check():
    '''runs internal checks to find errors'''
    _files_exist()



'''PaDRe -- Patient Data Repository

Library to organize participant imaging data'''

version = 2.0

_global_experiment = None
_include_all = False

from config import *

import sys
def error(msg):
	sys.stderr.write('## PaDRe Error: %s\n' % msg.strip())
	sys.stderr.flush()

import subject
from subject import Subject,subjects
import maint

from forgiving_dict import ForgivingDict

def load(s):
    return Subject.load(s)

def set_experiment(exp):
    global _global_experiment
    _global_experiment = exp


import os,subprocess
import filecmp
import re
import padre

data_dir = os.path.join(padre.padre_root,'Data')
subdirs = ['sessions','raw','analyses','clinical','other']

def max_diff(dset_a,dset_b):
    for dset in [dset_a,dset_b]:
        if not os.path.exists(dset):
            raise IOError('Could not find file: %s' % dset)
    try:
        with open(os.devnull,'w') as null:
            return float(subprocess.check_output(['3dBrickStat','-max','3dcalc( -a %s -b %s -expr abs(a-b) )' %(dset_a,dset_b)],stderr=null))
    except subprocess.CalledProcessError:
        return float('inf')

def make_subdirs(subj):
    for d in subdirs:
        full_d = os.path.join(data_dir,subj,d)
        if not os.path.exists(full_d):
            os.makedirs(full_d)

def dirs_from_tree(dirname):
    dirs = []
    for root, subdirs, files in os.walk(dirname):
        for d in subdirs:
            full_dir = os.path.join(root,d)
            if any([os.path.isfile(os.path.join(full_dir,x)) for x in os.listdir(full_dir)]):
                dirs.append(full_dir)
    return dirs

def dirs_equal(dirs):
    try:
        first_dir = dirs[0]
        other_dirs = list(dirs)
        del(other_dirs[other_dirs.index(first_dir)])
        for root, dirs, files in os.walk(first_dir):
            common_root = root[len(first_dir):]
            other_roots = [os.path.join(x,common_root) for x in other_dirs]
            this_content = os.listdir(root)
            other_contents = [os.listdir(x) for x in other_roots]
            if not all([len(this_content)==len(x) for x in other_contents]):
                return False
            for f in this_content:
                if not all([filecmp.cmp(os.path.join(root,f),os.path.join(x,f),False) for x in other_roots]):
                    return False
    except OSError:
        return False
    return True

label_map = {
    'storymath': ['story[\W_]*math'],
    'sdtones': ['s[md][\W_]*tones', 'fmri_tones', '[1-9]time'],
    'prep': ['3_plane_loc_', 'ASSET_CALIBRATION_', 'practice'],
    'anatomy': ['spgr', 'AX_3D_', 'anat'],
    'field_map': ['Ax_GRE'],
    'dti': ['dti'],
    'flair': ['flair'],
    'rest': ['rest'],
    'text': ['text_[0-9]'],
    'semphon': ['semphon'],
    'defname': ['def[\W_]name', 'dn\d', 'dn_run'],
    'aphname': ['aphname']
}

def categorize_dset(dset):
    standard_form = re.match(r'.*ep.\d+-\d+_\d+-(.*?)(\d+_reps)?(_\d+rows)?.nii.gz',dset)
    if standard_form:
        label = standard_form.group(1)
        for d in label_map:
            for m in label_map[d]:
                if re.search(m,label,re.IGNORECASE):
                    return d

def scan_subjects_dir():
    '''this no longer works'''
    raise Exception('This function has been depreciated and is not currently working')
    for subject_id in os.listdir(padre.data_dir):
        subj = padre.Subject.create(subject_id)
        for session in os.listdir(padre.sessions_subject_dir(subj)):
            if session not in subj.sessions:
                subj.new_session(session)
                try:
                    if str(int(session))==session:
                        subj._sessions[session]['date'] = session
                except ValueError:
                    pass
            already_found_dsets = []
            for set_type in subj._sessions[session]['sets']:
                already_found_dsets += subj._sessions[session]['sets'][set_type]
            for dset in sorted(os.listdir(os.path.join(padre.sessions_subject_dir(subj),session))):
                if dset not in already_found_dsets:
                    type_guess = categorize_dset(dset)
                    if type_guess:
                        if type_guess not in subj.sessions[session]['sets']:
                            subj.sessions[session]['sets'][type_guess] = []
                        subj.sessions[session]['sets'][type_guess].append(dset)
        subj.save()
    

def find_dup_sessions():
    ''' find sessions that are complete duplicates and delete them
    
    **Not complete**'''
    raise Exception('This function is not complete')
    for subj in p.subjects():
        for sess in subj.sessions:
            if sum([len(subj.sessions[sess]['labels'][x]) for x in subj.sessions[sess]['labels']])==0:
                continue
            matches = dict(subj.sessions)
            del(matches[sess])
            for label in subj.sessions[sess]['labels']:
                for dset in subj.sessions[sess]['labels'][label]:
                    loop_matches = list(matches)
                    for other_dir in loop_matches:
                        match = False
                        try:
                            for other_label in subj.sessions[other_dir]['labels']:
                                for other_dset in subj.sessions[other_dir]['labels'][other_label]:
                                    d = org.max_diff(os.path.join(p.sessions_subject_dir(subj),sess,dset),os.path.join(p.sessions_subject_dir(subj),other_dir,other_dset))
                                    if d<1.0:
                                        match = True
                                        raise ValueError
                        except ValueError:
                            pass
                        if not match:
                            del(matches[other_dir])
            if matches:
                print '%s: %s is duplicated within %s' % (subj,sess,matches.keys()[0])

if __name__=='__main__':
    import sys
    label = categorize_dset(sys.argv[1])
    print '%s: %s' % (label,sys.argv[1])
    
# find un-JSON'ed files
# create/update JSON with guesses
# print unknown labels
# scan raw for dups, del them
# autocreate each session
# label autocreated sessions


#!/usr/bin/env python
''' daemon to automatically collect and organize data into padre'''

import urllib2, re, json, os, shutil, socket, tarfile, tempfile, fnmatch
from dateutil.parser import parse
import neural as nl
import padre as p
from fuzzywuzzy import process

import_location = os.path.join(p.padre_root,'Import')
processed_location = os.path.join(p.padre_root,'Processed')
import_log_file = os.path.join(import_location,'import_log.json')

import imp
c = imp.load_source('padre_config',os.path.join(p.padre_root,'padre_config.py'))

import_log = {}
if not os.path.exists(import_location):
    os.makedirs(import_location)
if os.path.exists(import_log_file):
    with open(import_log_file) as f:
        import_log = json.loads(f.read())

def save_log():
    with open(import_log_file,'w') as f:
        f.write(json.dumps(import_log))    

def import_archive(full_file,subject_guess):
    tmp_dir = tempfile.mkdtemp()
    tmp_location = os.path.join(tmp_dir,'_tmp_unarchive')
    out_dir = os.path.join(processed_location,subject_guess)
    for d in [tmp_location,tmp_location + '-sorted']:
        if os.path.exists(d):
            shutil.rmtree(d)
    import_log[full_file] = {'modified':os.path.getmtime(full_file)}
    if nl.is_archive(full_file):
        with nl.notify('uncompressing files...'):
            os.makedirs(tmp_location)
            nl.unarchive(full_file,tmp_location)
    else:
        with nl.notify('copying files...'):
            shutil.copytree(full_file,tmp_location)
    with nl.notify('sorting files...'):
        nl.dicom.organize_dir(tmp_location)
    dsets_made = {}
    if os.path.exists(tmp_location + '-sorted'):
        if not os.path.exists(os.path.join(out_dir,'raw')):
            os.makedirs(os.path.join(out_dir,'raw'))
        with nl.run_in(tmp_location + '-sorted'):
            for subdir in os.listdir('.'):
                if 'dset_ignore_masks' in dir(c):
                    ignore = False
                    for m in c.dset_ignore_masks:
                        if m in subdir:
                            ignore = True
                    if ignore:
                        continue
                nl.notify('creating dataset from %s' % subdir)
                import_log[full_file][subdir] = {}
                if not nl.dicom.create_dset(subdir):
                    import_log[full_file][subdir]['error'] = True
                if os.path.exists(subdir + '.nii.gz'):
                    import_log[full_file][subdir]['complete'] = True
                    shutil.move(subdir + '.nii.gz',os.path.join(out_dir,subdir+'.nii.gz'))
                    session_name = subdir.split('-')[1]
                    if session_name not in dsets_made:
                        dsets_made[session_name] = []
                    dsets_made[session_name].append(os.path.join(out_dir,subdir+'.nii.gz'))
                if 'complete' in import_log[full_file][subdir] and import_log[full_file][subdir]['complete']:
                    if 'error' in import_log[full_file][subdir] and import_log[full_file][subdir]['error']:
                        nl.notify('created dataset %s, but Dimon returned an error' % (subdir+'.nii.gz'),nl.level.error)
                    else:
                        nl.notify('successfully created dataset %s' % (subdir+'.nii.gz'))
                else:
                    nl.notify('failed to create dataset from directory %s' % subdir,level=nl.level.error)
        nl.notify('moving raw data...')
        raw_data = []
        for subdir in os.listdir(tmp_location + '-sorted'):
            out_file = os.path.join(out_dir,'raw',subdir + '.tgz')
            if not os.path.exists(out_file):
                try:
                    with tarfile.open(out_file,'w:gz') as tgz:
                        tgz.add(os.path.join(tmp_location + '-sorted',subdir))
                except tarfile.TarError:
                    nl.notify('Error creating compressed raw directory %s' % out_file,level=nl.level.error)
                else:
                    raw_data.append(subdir + '.tgz')
        for r,ds,fs in os.walk(tmp_location):
            for f in fs:
                out_file = os.path.join(out_dir,'raw','unsorted',r.lstrip(tmp_location).lstrip('/'),f)
                if not os.path.exists(out_file):
                    if not os.path.exists(os.path.dirname(out_file)):
                        os.makedirs(os.path.dirname(out_file))
                    shutil.move(os.path.join(r,f),out_file)
                    raw_data.append(os.path.join('unsorted',r.lstrip(tmp_location).lstrip('/'),f))
        nl.notify('importing into padre...')
        for session in dsets_made:
            p.maint.import_to_padre(subject_guess,session,dsets_made[session],raw_data,out_dir)
    else:
        nl.notify('didn\'t find any files...')
    shutil.rmtree(tmp_dir)
    save_log()



def unpack_new_archives(pi):
    with nl.notify('Scanning PI %s for new archives' % pi):
        for root,dirs,files in os.walk(os.path.join(import_location,pi)):
            for fname in files + dirs:
                full_file = os.path.join(root,fname)
                if full_file not in import_log:
                    # Add in a check for the modification date
                    if any([fnmatch.fnmatch(fname,m) for m in c.archive_masks]):
                        with nl.notify('Found new archive "%s"' % full_file):
                            subject_guess = os.path.basename(os.path.dirname(full_file))
                            nl.notify('guessing the subject number is %s' % subject_guess)
                            import_archive(full_file,subject_guess)

def rsync_remote(pi):
    with nl.notify('rsync\'ing data for PI %s' % pi):
        nl.run([c.rsync,c.rsync_options,'-e','ssh -i %s' % c.server_id,'%s/%s' % (c.server_name,pi),import_location])

if __name__ == '__main__':
    for pi in c.pis:
        rsync_remote(pi)
        unpack_new_archives(pi)
                
            



from distutils.core import setup
setup(
  name = 'padre',
  packages = ['padre'],
  scripts = ['padre_demon.py','padre_web.py','padre_buddy.py'],
  data_files = [
      ('padre_web/views',[
          "views/header.stpl",
          "views/footer.stpl",
          "views/sidebar.stpl",
          "views/index.stpl",
          "views/list_subjects.stpl",
          "views/edit_session.stpl", 
          "views/edit_subject.stpl",
          "views/favicon.ico"
      ]),
      ('padre_web/views/style',[
          "views/style/back.png",
          "views/style/bullet.png",
          "views/style/datepicker.css",
          "views/style/footer.png",
          "views/style/graphic.png",
          "views/style/link.png",
          "views/style/search.png",
          "views/style/side_back.png",
          "views/style/side_base.png",
          "views/style/side_top.png",
          "views/style/style.css",
          "views/style/tab.png",
          "views/style/tab_selected.png"
      ])
  ],
  version = '2.1.3',
  description = 'participant data repository',
  long_description = '''PaDRe - a participant data repository library (in Python)
--------------------------------------------------------

This library is used to organize a large data repository of imaging data. The functions
consist of several helper functions that make accessing the data easier, and contain 
meta-data about subjects

Documentation: http://padre.readthedocs.org
''',
  author = 'Bill Gross',
  author_email = 'bill.gross@me.com',
  url = 'https://github.com/azraq27/padre',
  download_url = 'https://github.com/azraq27/padre/tarball/2.1.3',
  keywords = ['neuroimaging', 'afni', 'fsl', 'fmri'],
  classifiers = [
      'Topic :: Scientific/Engineering',
      'Intended Audience :: Science/Research',
      'Development Status :: 3 - Alpha'
  ],
  install_requires=[
      'neural-fmri',
      'pydicom',
      'fuzzywuzzy',
      'python-Levenshtein'
  ]
)


class ForgivingDict(dict):
    '''will return None instead of raising KeyError'''
    def __getitem__(self, key):
        try:
            return dict.__getitem__(self,key)
        except KeyError:
            return None
    
    @classmethod
    def copy_nested_dict(cls,orig_iter):
        '''will iterate through `orig_iter` and convert all dicts objects to ForgivingDict's'''
        if orig_iter.__class__ == dict:
            return_iter = cls()
        else:
            return_iter = []
        if isinstance(orig_iter,dict):
            for i in orig_iter:
                if '__iter__' in dir(orig_iter[i]):
                    return_iter[i] = cls.copy_nested_dict(orig_iter[i])
                else:
                    return_iter[i] = orig_iter[i]
        else:
            for i in orig_iter:
                if '__iter__' in dir(i):
                    return_iter.append(cls.copy_nested_dict(i))
                else:
                    return_iter.append(i)
        if orig_iter.__class__ != dict:
            return_iter = orig_iter.__class__(return_iter)
        return return_iter


#!/usr/bin/python
# depends on:
#   neural
#   padre
#   fuzzywuzzy
#   openpyxl

import argparse,os,json,sys,shutil
from fuzzywuzzy import process
import padre as p
import neural as nl
import openpyxl

fuzzyness = 90

listable_objects = {
    'atlases': ['atlases','templates'],
    'subjects': ['subjects','patients'],
    'experiments': ['experiments', 'studies', 'study'],
    'labels': ['labels','tasks'],
    'tags': ['types','tags'],
    'sessions': ['sessions','scans','dates'],
    'dsets': ['dsets','datasets','runs','files'],
    'meta': ['meta','metadata','behavioral','eprime']
}

p.subjects()
all_subjects = p.subject._all_subjects.keys()
all_labels = p.subject.tasks
all_tags = p.subject.tags
all_experiments = p.subject.experiments

all_objects = [(x,'subject') for x in all_subjects] + [(x,'label') for x in all_labels] + [(x,'tag') for x in all_tags] + [(x,'experiment') for x in all_experiments]

def error(msg,miss=None):
    nl.notify(msg + '\n',level=nl.level.error)
    if miss:
        try:
            import urllib,urllib2
            data = {'miss':miss}
            data_enc = urllib.urlencode(data)
            urllib2.urlopen('http://wolflion.org/cgi-bin/report.py?%s' % data_enc)
        except:
            pass
    sys.exit()

def identify_object(obj):
    '''tries to match to an existing object
    returns a tuple of (object,object_id)'''
    best_match = process.extractOne(obj,[x[0] for x in all_objects])
    if best_match and best_match[1]>fuzzyness:
        return [x for x in all_objects if x[0]==best_match[0]][0]
    else:
        return None

def identify_listable(keyword):
    '''tries to match keyword to some listable object'''    
    possible_names = [x for s in listable_objects.values() for x in s]
    best_match = process.extractOne(keyword,possible_names)
    if best_match[1]>fuzzyness:
        return [x for x in listable_objects if best_match[0] in listable_objects[x]][0]
    else:
        return None

def identify_dsets(dset_names,subjects):
    '''try to find each dset filename in ``dset_names`` within the list of subject objects ``subjects``'''
    found_dsets = {}
    leftover_names = list(dset_names)
    for subj in subjects:
        for sess in subj._sessions:
            for label in subj._sessions[sess]['labels']:
                for i in xrange(len(subj._sessions[sess]['labels'][label])):
                    dset = subj._sessions[sess]['labels'][label][i]
                    if dset['filename'] in dset_names:
                        if dset['filename'] in found_dsets:
                            error('Error: with the given parameters, I found the dataset name %s more than once' % dset['filename'])
                            return None
                        found_dsets[dset['filename']] = (subj,sess,label,dset,i)
                        del(leftover_names[leftover_names.index(dset['filename'])])
    return (found_dsets,leftover_names)

def identify_params(args,allow_unidentified=False):
    '''go through the args and try to assign unassigned jibberish'''
    params = {'subject':args.subject,'session':args.session,'tag':args.tag,'experiment':args.experiment,'label':args.label}
    global all_objects
    listable = None
    added_sessions = False
    unidentified_args = []
    for arg in args.something:
        if params['subject'] and not added_sessions:
            subj = p.load(params['subject'])
            if subj:
                all_objects += [(x,'session') for x in subj.sessions]
            added_sessions = True
        obj = identify_object(arg)
        if obj:
            params[obj[1]] = obj[0]
        else:
            l = identify_listable(arg)
            if l:
                listable = l
            else:
                unidentified_args.append(arg)
    if not allow_unidentified and len(unidentified_args)>0:
        error('Error: Sorry, I can\'t figure out what you mean by "%s"' % arg, arg)
    
    return (params,listable,unidentified_args)
 

def padre_info(args):
    pass

def subjects_from_params(params):
    if params['subject']:
        all_subjs = [p.load(params['subject'])]
    else:
        all_subjs = p.subjects(tag=params['tag'],label=params['label'],experiment=params['experiment'])
    return all_subjs

def padre_list(args):
    (params,listable,unidentified_args) = identify_params(args,True)
    
    if listable==None:
        error('Error: You didn\'t tell me what you wanted to list!')
        return
    
    all_subjs = subjects_from_params(params)
    
    all_dsets = {}
    if listable=='meta':
        if len(unidentified_args)>0:
            (all_dsets,unidentified_args) = identify_dsets(unidentified_args,all_subjs)
   
    if listable=='meta':
        if len(all_dsets)==0:
            error('Error: To list meta-data, you need to give me the name of a dataset')
            return
        for dset in all_dsets:
            subj,sess,label,dset_dict,i = all_dsets[dset]
            for meta in dset_dict['meta']:
                print '\t'.join([dset,meta,dset_dict['meta'][meta]])
    
    if listable=='atlases':
        if args.quiet:
            print '\n'.join(p.atlases.keys())
        else:
            for atlas in p.atlases:
                print atlas,
                if 'space' in p.atlases[atlas]:
                    print ' (%s)' % p.atlases[atlas]['space'],
                if 'notes' in p.atlases[atlas]:
                    print ': ' + p.atlases[atlas]['notes'],
                print ''
    if listable=='subjects':
        print '\n'.join([str(x) for x in all_subjs])
        return
    if listable=='labels':
        labels = set(nl.flatten([[x._sessions[y]['labels'].keys() for y in x._sessions] for x in all_subjs]))
        print '\n'.join(labels)
        return
    if listable=='experiments':
        experiments = set(nl.flatten([[x._sessions[y]['experiment'] for y in x._sessions if 'experiment' in x._sessions[y]] for x in all_subjs]))
        experiments = [x for x in experiments if x!=None and x!='']
        print '\n'.join(experiments)
        return
    if listable=='tags':
        tags = set(nl.flatten([[x._sessions[y]['tags'] for y in x._sessions if 'tags' in x._sessions[y]] for x in all_subjs]))
        tags = [x for x in tags if x!=None and x!='']
        print '\n'.join(tags)
        return
    if listable=='sessions':
        if not params['subject']:
            error('Error: sorry, I need a subject to list their sessions')
        else:
            subj = p.load(params['subject'])
            if subj:
                sessions = list(subj.sessions)
                sessions = [x for x in sessions if len(subj.dsets(session=x,tag=params['tag'],label=params['label'],experiment=params['experiment']))>0]
                print '\n'.join(sessions)
        return
    if listable=='dsets':
        if not params['subject']:
            error('Error: sorry, I need a subject to list their datasets')
        else:
            subj = p.load(params['subject'])
            if subj:
                dsets = subj.dsets(session=params['session'],tag=params['tag'],label=params['label'],experiment=params['experiment'])
                print '\n'.join([str(x) for x in dsets])
        return

def padre_link(args):
    (params,listable,unidentified_args) = identify_params(args)
    
    if not params['subject']:
        error('Error: You at least need to give me a subject!')
    subj = p.load(params['subject'])
    if not subj:
        error('Error: Couldn\'t load subject %s' % params['subject'])
    dsets = subj.dsets(session=params['session'],tag=params['tag'],label=params['label'],experiment=params['experiment'])
    for dset in dsets:
        os.symlink(str(dset),os.path.basename(str(dset)))

def padre_add_meta(args):
    (params,listable,unidentified_args) = identify_params(args,True)
    all_subjs = subjects_from_params(params)
    (all_dsets,unidentified_args) = identify_dsets(unidentified_args,all_subjs)
    if len(all_dsets)!=1:
        error('Error: To add meta-data, you need to give me the name of exactly one dataset')
    (subj,sess,label,dset,i) = all_dsets[0]
    subj._sessions[sess][label][i]['meta'][args.meta_type] = os.path.basename(args.filename)
    shutil.copy(args.filename,os.path.join(p.sessions_dir(subj),sess))
    print 'Added %s (%s) to %s' % (os.path.basename(args.filename),args.meta_type,str(subj))
 
if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--quiet','-q',action='store_true',help='just print the bare information (good for scripts)')
    
    specify_group = parser.add_argument_group(title='explicitly specify object',description='(certain commands are also able to parse fuzzy keyword arguments)')
    specify_group.add_argument('-s','--subject',help='subject id')
    specify_group.add_argument('-n','--session',help='session identifier, by convention date of the scanning session in the format YYYYMMDD, but could really be any unique string')
    specify_group.add_argument('-t','--tag',help='session tag')
    specify_group.add_argument('-l','--label',help='label for dataset (the type of dataset; anatomy, sdtones)')
    specify_group.add_argument('-e','--experiment',help='experiment identifier')
    subparsers = parser.add_subparsers(title='commands')
    
    parser_info = subparsers.add_parser('info',help='get details on a single object')
    parser_info.add_argument('something',nargs='*',help='keywords identifying what you want')
    parser_info.set_defaults(func=padre_info)
    
    parser_list = subparsers.add_parser('list',help='retrieve a list of things')
    parser_list.add_argument('something',nargs='*',help='keywords identifying all the things you want')
    parser_list.set_defaults(func=padre_list)
    
    parser_link = subparsers.add_parser('link',help='create symbolic links in the current directory to datasets matching the given parameters')
    parser_link.add_argument('something',nargs='*',help='keywords identifying the datasets you want')
    parser_link.set_defaults(func=padre_link)
    
    parser_add_meta = subparsers.add_parser('add_meta',help='attach a file as metadata to a dataset')
    parser_add_meta.add_argument('meta_type',help='keyword to identify the metadata file')
    parser_add_meta.add_argument('filename',help='filename of meta file you\'d like to add')
    parser_add_meta.add_argument('something',nargs='*',help='keywords identifying all the things you want')
    parser_add_meta.set_defaults(func=padre_add_meta)
    
    args = parser.parse_args()
    args.func(args)
